#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Gappy-POD Module
#
# Last rev: 28/02/2025

from .wrapper import GappyPOD
from .utils import set_random_elements_to_zero, delete_snapshot

del wrapper
