#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# SPOD Module
#
# Last rev: 06/10/2022

from .wrapper import run
from .utils   import extract_modes, save, load
from .plots import plotMode, plotSpectra
